/*
 * block.c --- iterate over all blocks in an inode
 * 
 * Copyright (C) 1993, 1994, 1995, 1996 Theodore Ts'o.
 *
 * %Begin-Header%
 * This file may be redistributed under the terms of the GNU Public
 * License.
 * %End-Header%
 */

#include <stdio.h>
#include <string.h>
#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "ext2_fs.h"
#include "ext2fs.h"

struct block_context {
	ext2_filsys	fs;
	int (*func)(ext2_filsys	fs,
		    blk_t	*blocknr,
		    e2_blkcnt_t	bcount,
		    blk_t	ref_blk,
		    int		ref_offset,
		    void	*priv_data);
	e2_blkcnt_t	bcount;
	int		bsize;
	int		flags;
	errcode_t	errcode;
	char	*ind_buf;
	char	*dind_buf;
	char	*tind_buf;
	void	*priv_data;
};

/* Begin extents.c */

#ifdef EXT_DEBUG
static void ext_show_header(struct ext3_extent_header *eh)
{
	printf("header: magic=%x entries=%u max=%u depth=%u generation=%u\n",
	       eh->eh_magic, eh->eh_entries, eh->eh_max, eh->eh_depth,
	       eh->eh_generation);
}

static void ext_show_index(struct ext3_extent_idx *ix)
{
	printf("index: block=%u leaf=%u leaf_hi=%u unused=%u\n",
	       ix->ei_block, ix->ei_leaf, ix->ei_leaf_hi, ix->ei_unused);
}

static void ext_show_extent(struct ext3_extent *ex)
{
	printf("extent: block=%u-%u len=%u start=%u start_hi=%u\n",
	       ex->ee_block, ex->ee_block + ex->ee_len - 1,
	       ex->ee_len, ex->ee_start, ex->ee_start_hi);
}

#define ext_printf(fmt, args...) printf(fmt, ## args)
#else
#define ext_show_header(eh) do { } while (0)
#define ext_show_index(ix) do { } while (0)
#define ext_show_extent(ex) do { } while (0)
#define ext_printf(fmt, args...) do { } while (0)
#endif

/* Internal function for ext2fs_block_iterate2() to recursively walk the
 * extent tree, with a callback function for each block.  We also call the
 * callback function on index blocks unless BLOCK_FLAG_DATA_ONLY is given.
 * We traverse the tree in-order (internal nodes before their children)
 * unless BLOCK_FLAG_DEPTH_FIRST is given.
 *
 * See also block_bmap_extents(). */
static int block_iterate_extents(void *eh_buf, unsigned bufsize, blk_t ref_block,
			  int ref_offset EXT2FS_ATTR((unused)),
			  struct block_context *ctx)
{
	struct ext3_extent_header *orig_eh, *eh;
	struct ext3_extent *ex, *ex_prev = NULL;
	int ret = 0;
	int item, offs, flags, split_flag = 0;
	blk_t  block_address;

	orig_eh = eh = eh_buf;

	if (ext2fs_extent_header_verify(eh, bufsize))
		return BLOCK_ERROR;

	if (eh->eh_depth == 0) {
		ex = EXT_FIRST_EXTENT(eh);
		for (item = 0; item < eh->eh_entries; item++, ex++) {
			ext_show_extent(ex);
			for (offs = 0; offs < ex->ee_len; offs++) {
				block_address = ex->ee_start + offs;
				flags = (*ctx->func)(ctx->fs, &block_address,
						     (ex->ee_block + offs),
						     ref_block, item,
						     ctx->priv_data);
				if (flags & (BLOCK_ABORT | BLOCK_ERROR)) {
					ret |= flags &(BLOCK_ABORT|BLOCK_ERROR);
					return ret;
				}
				if (!(flags & BLOCK_CHANGED))
					continue;

				ext_printf("extent leaf changed: "
					   "block was %u+%u = %u, now %u\n",
					   ex->ee_start, offs,
					   ex->ee_start + offs, block_address);

				/* FIXME: 48-bit support */
				if (ex_prev &&
				    block_address ==
				    ex_prev->ee_start + ex_prev->ee_len &&
				    ex->ee_block + offs ==
				    ex_prev->ee_block + ex_prev->ee_len) {
					/* can merge block with prev extent */
					ex_prev->ee_len++;
					ex->ee_len--;
					ret |= BLOCK_CHANGED;

					if (ex->ee_len == 0) {
						/* no blocks left in this one */
						ext2fs_extent_remove(eh, ex);
						item--; ex--;
						break;
					} else {
						/* XXX 48-bit */
						ex->ee_start++;
						ex->ee_block++;
						offs--;
					}

				} else if (offs > 0 && /* implies ee_len > 1 */
					   (ctx->errcode =
					    ext2fs_extent_split(ctx->fs, &eh,
								&ex, offs,
								&split_flag)
					    /* advance ex past newly split item,
					     * comparison is bogus to make sure
					     * increment doesn't change logic */
					    || (offs > 0 && ex++ == NULL))) {
					/* split before new block failed */
					ret |= BLOCK_ABORT | BLOCK_ERROR;
					return ret;

				} else if (ex->ee_len > 1 &&
					   (ctx->errcode =
					    ext2fs_extent_split(ctx->fs, &eh,
								&ex, 1,
								&split_flag))) {
					/* split after new block failed */
					ret |= BLOCK_ABORT | BLOCK_ERROR;
					return ret;

				} else {
					if (ex->ee_len != 1) {
						/* this is an internal error */
						ctx->errcode =
						       EXT2_ET_EXTENT_INDEX_BAD;
						ret |= BLOCK_ABORT |BLOCK_ERROR;
						return ret;
					}
					/* FIXME: 48-bit */
					ex->ee_start = block_address;
					ret |= BLOCK_CHANGED;
				}
			}
			ex_prev = ex;
		}
		/* Multi level split at depth == 0.
		 * ex has been changed to point to  newly allocated block
		 * buffer. And after returning  in this scenario, only inode is
		 * updated with changed i_block. Hence explicitly write to the
		 * block is required. */
		if (split_flag == BLOCK_CHANGED) {
			struct ext3_extent_idx *ix = EXT_FIRST_INDEX(orig_eh);
			ctx->errcode = ext2fs_write_ext_block(ctx->fs,
							      ix->ei_leaf, eh);
		}
	} else {
		char *block_buf;
		struct ext3_extent_idx *ix;

		ret = ext2fs_get_mem(ctx->fs->blocksize, &block_buf);
		if (ret)
			return ret;

		ext_show_header(eh);
		ix = EXT_FIRST_INDEX(eh);
		for (item = 0; item < eh->eh_entries; item++, ix++) {
			ext_show_index(ix);
			/* index is processed first in e2fsck case */
			if (!(ctx->flags & BLOCK_FLAG_DEPTH_TRAVERSE) &&
			    !(ctx->flags & BLOCK_FLAG_DATA_ONLY)) {
				block_address = ix->ei_leaf;
				flags = (*ctx->func)(ctx->fs, &block_address,
						     BLOCK_COUNT_IND, ref_block,
						     item, ctx->priv_data);
				if (flags & (BLOCK_ABORT | BLOCK_ERROR)) {
					ret |= flags &(BLOCK_ABORT|BLOCK_ERROR);
					goto free_buf;
				}
				if (flags & BLOCK_CHANGED) {
					/* index has no more block, remove it */
					/* FIXME: 48-bit */
					ix->ei_leaf = block_address;
					if (ix->ei_leaf == 0 &&
					    ix->ei_leaf_hi == 0 &&
					    ext2fs_extent_index_remove(eh, ix)){
						ret |= BLOCK_ABORT |BLOCK_ERROR;
						goto free_buf;
					}

					ret |= BLOCK_CHANGED;
					if (ix->ei_leaf == 0 &&
					    ix->ei_leaf_hi == 0) {
						--item; --ix;
						continue;
					}
					/* remapped? */
				}
			}
			ctx->errcode = ext2fs_read_ext_block(ctx->fs,
							     ix->ei_leaf,
							     block_buf);
			if (ctx->errcode) {
				ret |= BLOCK_ERROR;
				goto free_buf;
			}
			flags = block_iterate_extents(block_buf,
						      ctx->fs->blocksize,
						      ix->ei_leaf, item, ctx);
			if (flags & BLOCK_CHANGED) {
				struct ext3_extent_header *nh;
				ctx->errcode =
					ext2fs_write_ext_block(ctx->fs,
							       ix->ei_leaf,
							       block_buf);

				nh = (struct ext3_extent_header *)block_buf;
				if (nh->eh_entries == 0)
					ix->ei_leaf = ix->ei_leaf_hi = 0;
			}
			if (flags & (BLOCK_ABORT | BLOCK_ERROR)) {
				ret |= flags & (BLOCK_ABORT | BLOCK_ERROR);
				goto free_buf;
			}
			if ((ctx->flags & BLOCK_FLAG_DEPTH_TRAVERSE) &&
			    !(ctx->flags & BLOCK_FLAG_DATA_ONLY)) {
				flags = (*ctx->func)(ctx->fs, &block_address,
						     BLOCK_COUNT_IND, ref_block,
						     item, ctx->priv_data);
				if (flags & (BLOCK_ABORT | BLOCK_ERROR)) {
					ret |= flags &(BLOCK_ABORT|BLOCK_ERROR);
					goto free_buf;
				}
				if (flags & BLOCK_CHANGED)
					/* FIXME: 48-bit */
					ix->ei_leaf = block_address;
			}

			if (flags & BLOCK_CHANGED) {
				/* index has no more block, remove it */
				if (ix->ei_leaf == 0 && ix->ei_leaf_hi == 0 &&
				    ext2fs_extent_index_remove(eh, ix)) {
					ret |= BLOCK_ABORT |BLOCK_ERROR;
					goto free_buf;
				}

				ret |= BLOCK_CHANGED;
				if (ref_block == 0) {
					--item; --ix;
					continue;
				}
				/* remapped? */
			}
		}

	free_buf:
		ext2fs_free_mem(&block_buf);
	}
	return ret;
}



/* End extents.c */

#ifdef EXT_DEBUG
void ext_show_inode(struct ext2_inode *inode, ext2_ino_t ino)
{
	printf("inode: %u blocks: %u\n",
	       ino, inode->i_blocks);
}
#else
#define ext_show_inode(inode, ino) do { } while (0)
#endif

static int block_iterate_ind(blk_t *ind_block, blk_t ref_block,
			     int ref_offset, struct block_context *ctx)
{
	int	ret = 0, changed = 0;
	int	i, flags, limit, offset;
	blk_t	*block_nr;

	limit = ctx->fs->blocksize >> 2;
	if (!(ctx->flags & BLOCK_FLAG_DEPTH_TRAVERSE) &&
	    !(ctx->flags & BLOCK_FLAG_DATA_ONLY))
		ret = (*ctx->func)(ctx->fs, ind_block,
				   BLOCK_COUNT_IND, ref_block,
				   ref_offset, ctx->priv_data);
	if (!*ind_block || (ret & BLOCK_ABORT)) {
		ctx->bcount += limit;
		return ret;
	}
	if (*ind_block >= ctx->fs->super->s_blocks_count ||
	    *ind_block < ctx->fs->super->s_first_data_block) {
		ctx->errcode = EXT2_ET_BAD_IND_BLOCK;
		ret |= BLOCK_ERROR;
		return ret;
	}
	ctx->errcode = ext2fs_read_ind_block(ctx->fs, *ind_block, 
					     ctx->ind_buf);
	if (ctx->errcode) {
		ret |= BLOCK_ERROR;
		return ret;
	}

	block_nr = (blk_t *) ctx->ind_buf;
	offset = 0;
	if (ctx->flags & BLOCK_FLAG_APPEND) {
		for (i = 0; i < limit; i++, ctx->bcount++, block_nr++) {
			flags = (*ctx->func)(ctx->fs, block_nr, ctx->bcount,
					     *ind_block, offset, 
					     ctx->priv_data);
			changed	|= flags;
			if (flags & BLOCK_ABORT) {
				ret |= BLOCK_ABORT;
				break;
			}
			offset += sizeof(blk_t);
		}
	} else {
		for (i = 0; i < limit; i++, ctx->bcount++, block_nr++) {
			if (*block_nr == 0)
				continue;
			flags = (*ctx->func)(ctx->fs, block_nr, ctx->bcount,
					     *ind_block, offset, 
					     ctx->priv_data);
			changed	|= flags;
			if (flags & BLOCK_ABORT) {
				ret |= BLOCK_ABORT;
				break;
			}
			offset += sizeof(blk_t);
		}
	}
	if (changed & BLOCK_CHANGED) {
		ctx->errcode = ext2fs_write_ind_block(ctx->fs, *ind_block,
						      ctx->ind_buf);
		if (ctx->errcode)
			ret |= BLOCK_ERROR | BLOCK_ABORT;
	}
	if ((ctx->flags & BLOCK_FLAG_DEPTH_TRAVERSE) &&
	    !(ctx->flags & BLOCK_FLAG_DATA_ONLY) &&
	    !(ret & BLOCK_ABORT))
		ret |= (*ctx->func)(ctx->fs, ind_block,
				    BLOCK_COUNT_IND, ref_block,
				    ref_offset, ctx->priv_data);
	return ret;
}
	
static int block_iterate_dind(blk_t *dind_block, blk_t ref_block,
			      int ref_offset, struct block_context *ctx)
{
	int	ret = 0, changed = 0;
	int	i, flags, limit, offset;
	blk_t	*block_nr;

	limit = ctx->fs->blocksize >> 2;
	if (!(ctx->flags & (BLOCK_FLAG_DEPTH_TRAVERSE |
			    BLOCK_FLAG_DATA_ONLY)))
		ret = (*ctx->func)(ctx->fs, dind_block,
				   BLOCK_COUNT_DIND, ref_block,
				   ref_offset, ctx->priv_data);
	if (!*dind_block || (ret & BLOCK_ABORT)) {
		ctx->bcount += limit*limit;
		return ret;
	}
	if (*dind_block >= ctx->fs->super->s_blocks_count ||
	    *dind_block < ctx->fs->super->s_first_data_block) {
		ctx->errcode = EXT2_ET_BAD_DIND_BLOCK;
		ret |= BLOCK_ERROR;
		return ret;
	}
	ctx->errcode = ext2fs_read_ind_block(ctx->fs, *dind_block, 
					     ctx->dind_buf);
	if (ctx->errcode) {
		ret |= BLOCK_ERROR;
		return ret;
	}

	block_nr = (blk_t *) ctx->dind_buf;
	offset = 0;
	if (ctx->flags & BLOCK_FLAG_APPEND) {
		for (i = 0; i < limit; i++, block_nr++) {
			flags = block_iterate_ind(block_nr,
						  *dind_block, offset,
						  ctx);
			changed |= flags;
			if (flags & (BLOCK_ABORT | BLOCK_ERROR)) {
				ret |= flags & (BLOCK_ABORT | BLOCK_ERROR);
				break;
			}
			offset += sizeof(blk_t);
		}
	} else {
		for (i = 0; i < limit; i++, block_nr++) {
			if (*block_nr == 0) {
				ctx->bcount += limit;
				continue;
			}
			flags = block_iterate_ind(block_nr,
						  *dind_block, offset,
						  ctx);
			changed |= flags;
			if (flags & (BLOCK_ABORT | BLOCK_ERROR)) {
				ret |= flags & (BLOCK_ABORT | BLOCK_ERROR);
				break;
			}
			offset += sizeof(blk_t);
		}
	}
	if (changed & BLOCK_CHANGED) {
		ctx->errcode = ext2fs_write_ind_block(ctx->fs, *dind_block,
						      ctx->dind_buf);
		if (ctx->errcode)
			ret |= BLOCK_ERROR | BLOCK_ABORT;
	}
	if ((ctx->flags & BLOCK_FLAG_DEPTH_TRAVERSE) &&
	    !(ctx->flags & BLOCK_FLAG_DATA_ONLY) &&
	    !(ret & BLOCK_ABORT))
		ret |= (*ctx->func)(ctx->fs, dind_block,
				    BLOCK_COUNT_DIND, ref_block,
				    ref_offset, ctx->priv_data);
	return ret;
}
	
static int block_iterate_tind(blk_t *tind_block, blk_t ref_block,
			      int ref_offset, struct block_context *ctx)
{
	int	ret = 0, changed = 0;
	int	i, flags, limit, offset;
	blk_t	*block_nr;

	limit = ctx->fs->blocksize >> 2;
	if (!(ctx->flags & (BLOCK_FLAG_DEPTH_TRAVERSE |
			    BLOCK_FLAG_DATA_ONLY)))
		ret = (*ctx->func)(ctx->fs, tind_block,
				   BLOCK_COUNT_TIND, ref_block,
				   ref_offset, ctx->priv_data);
	if (!*tind_block || (ret & BLOCK_ABORT)) {
		ctx->bcount += limit*limit*limit;
		return ret;
	}
	if (*tind_block >= ctx->fs->super->s_blocks_count ||
	    *tind_block < ctx->fs->super->s_first_data_block) {
		ctx->errcode = EXT2_ET_BAD_TIND_BLOCK;
		ret |= BLOCK_ERROR;
		return ret;
	}
	ctx->errcode = ext2fs_read_ind_block(ctx->fs, *tind_block, 
					     ctx->tind_buf);
	if (ctx->errcode) {
		ret |= BLOCK_ERROR;
		return ret;
	}

	block_nr = (blk_t *) ctx->tind_buf;
	offset = 0;
	if (ctx->flags & BLOCK_FLAG_APPEND) {
		for (i = 0; i < limit; i++, block_nr++) {
			flags = block_iterate_dind(block_nr,
						   *tind_block,
						   offset, ctx);
			changed |= flags;
			if (flags & (BLOCK_ABORT | BLOCK_ERROR)) {
				ret |= flags & (BLOCK_ABORT | BLOCK_ERROR);
				break;
			}
			offset += sizeof(blk_t);
		}
	} else {
		for (i = 0; i < limit; i++, block_nr++) {
			if (*block_nr == 0) {
				ctx->bcount += limit*limit;
				continue;
			}
			flags = block_iterate_dind(block_nr,
						   *tind_block,
						   offset, ctx);
			changed |= flags;
			if (flags & (BLOCK_ABORT | BLOCK_ERROR)) {
				ret |= flags & (BLOCK_ABORT | BLOCK_ERROR);
				break;
			}
			offset += sizeof(blk_t);
		}
	}
	if (changed & BLOCK_CHANGED) {
		ctx->errcode = ext2fs_write_ind_block(ctx->fs, *tind_block,
						      ctx->tind_buf);
		if (ctx->errcode)
			ret |= BLOCK_ERROR | BLOCK_ABORT;
	}
	if ((ctx->flags & BLOCK_FLAG_DEPTH_TRAVERSE) &&
	    !(ctx->flags & BLOCK_FLAG_DATA_ONLY) &&
	    !(ret & BLOCK_ABORT))
		ret |= (*ctx->func)(ctx->fs, tind_block,
				    BLOCK_COUNT_TIND, ref_block,
				    ref_offset, ctx->priv_data);
	
	return ret;
}
	
errcode_t ext2fs_block_iterate2(ext2_filsys fs,
				ext2_ino_t ino,
				int	flags,
				char *block_buf,
				int (*func)(ext2_filsys fs,
					    blk_t	*blocknr,
					    e2_blkcnt_t	blockcnt,
					    blk_t	ref_blk,
					    int		ref_offset,
					    void	*priv_data),
				void *priv_data)
{
	int	i;
	int	ret = 0;
	blk_t	blocks[EXT2_N_BLOCKS];	/* directory data blocks */
	struct ext2_inode inode;
	errcode_t	retval;
	struct block_context ctx;
	int	limit;

	EXT2_CHECK_MAGIC(fs, EXT2_ET_MAGIC_EXT2FS_FILSYS);

	ctx.errcode = ext2fs_read_inode(fs, ino, &inode);
	if (ctx.errcode)
		return ctx.errcode;

	/*
	 * Check to see if we need to limit large files
	 */
	if (flags & BLOCK_FLAG_NO_LARGE) {
		if (!LINUX_S_ISDIR(inode.i_mode) &&
		    (inode.i_size_high != 0))
			return EXT2_ET_FILE_TOO_BIG;
	}

	/* The in-memory inode may have been changed by e2fsck */
	retval = ext2fs_get_blocks(fs, ino, blocks);
	if (retval)
		return retval;

	limit = fs->blocksize >> 2;

	ctx.fs = fs;
	ctx.func = func;
	ctx.priv_data = priv_data;
	ctx.flags = flags;
	ctx.bcount = 0;
	if (block_buf) {
		ctx.ind_buf = block_buf;
	} else {
		retval = ext2fs_get_mem(fs->blocksize * 3, &ctx.ind_buf);
		if (retval)
			return retval;
	}
	ctx.dind_buf = ctx.ind_buf + fs->blocksize;
	ctx.tind_buf = ctx.dind_buf + fs->blocksize;

	/*
	 * Iterate over the HURD translator block (if present)
	 */
	if ((fs->super->s_creator_os == EXT2_OS_HURD) &&
	    !(flags & BLOCK_FLAG_DATA_ONLY)) {
		if (inode.osd1.hurd1.h_i_translator) {
			ret |= (*ctx.func)(fs,
					   &inode.osd1.hurd1.h_i_translator,
					   BLOCK_COUNT_TRANSLATOR,
					   0, 0, priv_data);
			if (ret & BLOCK_ABORT)
				goto abort_exit;
		}
	}

	/* Iterate over normal data blocks with extents.
	 * We can't do any fixing here because this gets called by other
	 * callers than e2fsck_pass1->check_blocks(). */
	if (inode.i_flags & EXT4_EXTENTS_FL) {
		ext_show_inode(&inode, ino);
		ret |= block_iterate_extents(blocks, sizeof(blocks), 0, 0,&ctx);
		goto abort_exit;
	}

	/*
	 * Iterate over normal data blocks
	 */
	for (i = 0; i < EXT2_NDIR_BLOCKS ; i++, ctx.bcount++) {
		if (blocks[i] || (flags & BLOCK_FLAG_APPEND)) {
			ret |= (*ctx.func)(fs, &blocks[i],
					    ctx.bcount, 0, i, priv_data);
			if (ret & BLOCK_ABORT)
				goto abort_exit;
		}
	}
	if (*(blocks + EXT2_IND_BLOCK) || (flags & BLOCK_FLAG_APPEND)) {
		ret |= block_iterate_ind(blocks + EXT2_IND_BLOCK,
					 0, EXT2_IND_BLOCK, &ctx);
		if (ret & BLOCK_ABORT)
			goto abort_exit;
	} else
		ctx.bcount += limit;
	if (*(blocks + EXT2_DIND_BLOCK) || (flags & BLOCK_FLAG_APPEND)) {
		ret |= block_iterate_dind(blocks + EXT2_DIND_BLOCK,
					  0, EXT2_DIND_BLOCK, &ctx);
		if (ret & BLOCK_ABORT)
			goto abort_exit;
	} else
		ctx.bcount += limit * limit;
	if (*(blocks + EXT2_TIND_BLOCK) || (flags & BLOCK_FLAG_APPEND)) {
		ret |= block_iterate_tind(blocks + EXT2_TIND_BLOCK,
					  0, EXT2_TIND_BLOCK, &ctx);
		if (ret & BLOCK_ABORT)
			goto abort_exit;
	}

abort_exit:
	if (ret & BLOCK_CHANGED) {
		for (i=0; i < EXT2_N_BLOCKS; i++)
			inode.i_block[i] = blocks[i];
		retval = ext2fs_write_inode(fs, ino, &inode);
		if (retval)
			return retval;
	}

	if (!block_buf)
		ext2fs_free_mem(&ctx.ind_buf);

	return (ret & BLOCK_ERROR) ? ctx.errcode : 0;
}

/*
 * Emulate the old ext2fs_block_iterate function!
 */

struct xlate {
	int (*func)(ext2_filsys	fs,
		    blk_t	*blocknr,
		    int		bcount,
		    void	*priv_data);
	void *real_private;
};

#ifdef __TURBOC__
 #pragma argsused
#endif
static int xlate_func(ext2_filsys fs, blk_t *blocknr, e2_blkcnt_t blockcnt,
		      blk_t ref_block EXT2FS_ATTR((unused)),
		      int ref_offset EXT2FS_ATTR((unused)),
		      void *priv_data)
{
	struct xlate *xl = (struct xlate *) priv_data;

	return (*xl->func)(fs, blocknr, (int) blockcnt, xl->real_private);
}

errcode_t ext2fs_block_iterate(ext2_filsys fs,
			       ext2_ino_t ino,
			       int	flags,
			       char *block_buf,
			       int (*func)(ext2_filsys fs,
					   blk_t	*blocknr,
					   int	blockcnt,
					   void	*priv_data),
			       void *priv_data)
{
	struct xlate xl;
	
	xl.real_private = priv_data;
	xl.func = func;

	return ext2fs_block_iterate2(fs, ino, BLOCK_FLAG_NO_LARGE | flags,
				     block_buf, xlate_func, &xl);
}

