/*
 * extent.c --- iterate over all extents in an inode
 *
 * Copyright (C) 2006 Theodore Ts'o.
 *
 * %Begin-Header%
 * This file may be redistributed under the terms of the GNU Public
 * License.
 * %End-Header%
 */

#include <stdio.h>
#include <string.h>
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <errno.h>

#ifdef DEBUG
#define EXT_DEBUG
#endif

#include "ext2_fs.h"
#include "ext2fs.h"

struct block_context {
	ext2_filsys	fs;
	int (*func)(ext2_filsys	fs,
		    blk_t	*blocknr,
		    e2_blkcnt_t	bcount,
		    blk_t	ref_blk,
		    int		ref_offset,
		    void	*priv_data);
	e2_blkcnt_t	bcount;
	int		bsize;
	int		flags;
	errcode_t	errcode;
	char	*ind_buf;
	char	*dind_buf;
	char	*tind_buf;
	void	*priv_data;
};


#ifdef EXT_DEBUG
static void ext_show_header(struct ext3_extent_header *eh)
{
	printf("header: magic=%x entries=%u max=%u depth=%u generation=%u\n",
	       eh->eh_magic, eh->eh_entries, eh->eh_max, eh->eh_depth,
	       eh->eh_generation);
}

static void ext_show_index(struct ext3_extent_idx *ix)
{
	printf("index: block=%u leaf=%u leaf_hi=%u unused=%u\n",
	       ix->ei_block, ix->ei_leaf, ix->ei_leaf_hi, ix->ei_unused);
}

static void ext_show_extent(struct ext3_extent *ex)
{
	printf("extent: block=%u-%u len=%u start=%u start_hi=%u\n",
	       ex->ee_block, ex->ee_block + ex->ee_len - 1,
	       ex->ee_len, ex->ee_start, ex->ee_start_hi);
}

#define ext_printf(fmt, args...) printf(fmt, ## args)
#else
#define ext_show_header(eh) do { } while (0)
#define ext_show_index(ix) do { } while (0)
#define ext_show_extent(ex) do { } while (0)
#define ext_printf(fmt, args...) do { } while (0)
#endif

errcode_t ext2fs_extent_header_verify(struct ext3_extent_header *eh, int size)
{
	int eh_max, entry_size;

	ext_show_header(eh);
	if (eh->eh_magic != EXT3_EXT_MAGIC)
		return EXT2_ET_EXTENT_HEADER_BAD;
	if (eh->eh_entries > eh->eh_max)
		return EXT2_ET_EXTENT_HEADER_BAD;
	if (eh->eh_depth == 0)
		entry_size = sizeof(struct ext3_extent);
	else
		entry_size = sizeof(struct ext3_extent_idx);

	eh_max = (size - sizeof(*eh)) / entry_size;
	/* Allow two extent-sized items at the end of the block, for
	 * ext4_extent_tail with checksum in the future. */
	if (eh->eh_max > eh_max || eh->eh_max < eh_max - 2)
		return EXT2_ET_EXTENT_HEADER_BAD;

	return 0;
}

/* Verify that a single extent @ex is valid.  If @ex_prev is passed in,
 * then this was the previous logical extent in this block and we can
 * do additional sanity checking (though in case of error we don't know
 * which of the two extents is bad).  Similarly, if @ix is passed in
 * we can check that this extent is logically part of the index that
 * refers to it (though again we can't know which of the two is bad). */
errcode_t ext2fs_extent_verify(ext2_filsys fs, struct ext3_extent *ex,
			       struct ext3_extent *ex_prev,
			       struct ext3_extent_idx *ix, int ix_len)
{
	ext_show_extent(ex);
	/* FIXME: 48-bit support */
	if (ex->ee_start > fs->super->s_blocks_count)
		return EXT2_ET_EXTENT_LEAF_BAD;

	if (ex->ee_len == 0)
		return EXT2_ET_EXTENT_LEAF_BAD;

	if (ex->ee_len >= fs->super->s_blocks_per_group)
		return EXT2_ET_EXTENT_LEAF_BAD;

	if (ex_prev) {
		/* We can't have a zero logical block except for first index */
		if (ex->ee_block == 0)
			return EXT2_ET_EXTENT_LEAF_BAD;

		/* FIXME: 48-bit support */
		/* extents must be in logical offset order */
		if (ex->ee_block < ex_prev->ee_block + ex_prev->ee_len)
			return EXT2_ET_EXTENT_LEAF_BAD;

		/* extents must not overlap physical blocks */
		if ((ex->ee_start < ex_prev->ee_start + ex_prev->ee_len) &&
		    (ex->ee_start + ex->ee_len > ex_prev->ee_start))
			return EXT2_ET_EXTENT_LEAF_BAD;
	}

	if (ix) {
		/* FIXME: 48-bit support */
		if (ex->ee_block < ix->ei_block)
			return EXT2_ET_EXTENT_LEAF_BAD;

		if (ix_len && ex->ee_block + ex->ee_len > ix->ei_block + ix_len)
			return EXT2_ET_EXTENT_LEAF_BAD;
	}

	return 0;
}

errcode_t ext2fs_extent_index_verify(ext2_filsys fs,
				     struct ext3_extent_idx *ix,
				     struct ext3_extent_idx *ix_prev)
{
	ext_show_index(ix);
	/* FIXME: 48-bit support */
	if (ix->ei_leaf > fs->super->s_blocks_count)
		return EXT2_ET_EXTENT_INDEX_BAD;

	if (ix_prev == NULL)
		return 0;

	/* We can't have a zero logical block except for first index */
	if (ix->ei_block == 0)
		return EXT2_ET_EXTENT_INDEX_BAD;

	if (ix->ei_block <= ix_prev->ei_block)
		return EXT2_ET_EXTENT_INDEX_BAD;

	return 0;
}

errcode_t ext2fs_extent_remove(struct ext3_extent_header *eh,
			       struct ext3_extent *ex)
{
	int offs = ex - EXT_FIRST_EXTENT(eh);

	if (offs < 0 || offs > eh->eh_entries)
		return EXT2_ET_EXTENT_LEAF_BAD;

	ext_printf("remove extent: offset %u\n", offs);

	memmove(ex, ex + 1, (eh->eh_entries - offs - 1) * sizeof(*ex));
	--eh->eh_entries;

	return 0;
}

errcode_t ext2fs_extent_split_internal(struct ext3_extent_header *eh,
				       struct ext3_extent *ex, int offs)
{
	int entry = ex - EXT_FIRST_EXTENT(eh);
	struct ext3_extent *ex_new = ex + 1;

	ext_printf("split: ee_len: %u ee_block: %u ee_start: %u offset: %u\n",
		   ex->ee_len, ex->ee_block, ex->ee_start, offs);
	memmove(ex_new, ex, (eh->eh_entries - entry) * sizeof(*ex));
	++eh->eh_entries;

	ex->ee_len = offs;
	/* FIXME: 48-bit support */
	ex_new->ee_len -= offs;
	ex_new->ee_block += offs;
	ex_new->ee_start += offs;

	return 0;
}

errcode_t ext2fs_extent_split(ext2_filsys fs,
			      struct ext3_extent_header **eh_orig,
			      struct ext3_extent **ex_orig, int offs, int *flag)
{
	struct ext3_extent_header *eh_parent = *eh_orig;
	int retval, entry = *ex_orig - EXT_FIRST_EXTENT(eh_parent);
	blk_t new_block;
	char *buf;
	struct ext3_extent_idx *ei = EXT_FIRST_INDEX(eh_parent);

	if (entry < 0 || entry > (*eh_orig)->eh_entries)
		return EXT2_ET_EXTENT_LEAF_BAD;

	if (offs > (*ex_orig)->ee_len)
		return EXT2_ET_EXTENT_LEAF_BAD;

	if (eh_parent->eh_entries >= eh_parent->eh_max) {
		ext_printf("split: eh_entries: %u eh_max: %u\n",
			   eh_parent->eh_entries, eh_parent->eh_max);
		if (eh_parent->eh_max == 4) {
			struct ext3_extent_header *eh_child;
			struct ext3_extent *ex_child;

			retval = ext2fs_get_mem(fs->blocksize, &buf);

			if (retval)
				return EXT2_ET_EXTENT_NO_SPACE;

			memset(buf, 0, fs->blocksize);
			memcpy(buf, eh_parent, sizeof(*eh_parent) +
			       eh_parent->eh_entries * sizeof(*ex_child));
			eh_child = (struct ext3_extent_header *)buf;

			eh_child->eh_max = (fs->blocksize -
					    sizeof(struct ext3_extent_header)) /
					   sizeof(struct ext3_extent);
			retval = ext2fs_new_block(fs, (*ex_orig)->ee_block, 0,
						  &new_block);
			if (retval)
				return EXT2_ET_EXTENT_NO_SPACE;

			retval = io_channel_write_blk(fs->io, new_block, 1,buf);
			if (retval)
				return EXT2_ET_EXTENT_NO_SPACE;

			eh_parent->eh_entries = 1;
			eh_parent->eh_depth = 1;

			ex_child = EXT_FIRST_EXTENT(eh_child);
			ei->ei_block = ex_child->ee_block;
			/*XXX 48 bit support*/
			ei->ei_leaf = new_block;

			*eh_orig = eh_child;
			*ex_orig = EXT_FIRST_EXTENT(eh_child)  +
				sizeof(struct ext3_extent) * entry;

			*flag = BLOCK_CHANGED;
		} else {
			return EXT2_ET_EXTENT_NO_SPACE;
		}
	}

	return ext2fs_extent_split_internal(*eh_orig, *ex_orig, offs);
}

errcode_t ext2fs_extent_index_remove(struct ext3_extent_header *eh,
				     struct ext3_extent_idx *ix)
{
	struct ext3_extent_idx *first = EXT_FIRST_INDEX(eh);
	int offs = ix - first;

	ext_printf("remove index: offset %u\n", offs);

	memmove(ix, ix + 1, (eh->eh_entries - offs - 1) * sizeof(*ix));
	--eh->eh_entries;

	return 0;
}



/* Internal function for ext2fs_block_iterate().
 * See also block_bmap_extents(). */
static int block_iterate_extents(ext2_filsys fs,
				 struct ext3_extent_header *eh,
				 blk_t ref_block,
				 int	flags,
				 int (*func)(ext2_filsys fs,
					     struct ext2fs_extent *extent,
					     void	*priv_data),
				 int (*meta_func)(ext2_filsys fs,
						  blk64_t blk,
						  int blk_type,
						  char *buf,
						  void *priv_data),
				 void *priv_data)
{
	int ret_flags = 0;
	int i, j, ret;
	blk_t  block_address;
	errcode_t retval;
	struct ext3_extent *ex, *ex_prev = NULL;
	struct ext2fs_extent extent, extent_prev;

	printf("-->Entering block_iterate_extents depth %d\n", eh->eh_depth);
	if (eh->eh_depth == 0) {
		ex = EXT_FIRST_EXTENT(eh);
		for (i = 0; i < eh->eh_entries; i++, ex++) {
			ext_show_extent(ex);
			extent.e_lblk = ex->ee_block;
			extent.e_pblk = ex->ee_start +
				((long long) ex->ee_start_hi << 32);
			extent.e_len = ex->ee_len;
			ret_flags = func(fs, &extent, priv_data);
			if (ret_flags & BLOCK_ABORT) {
				ret |= BLOCK_ABORT;
				return ret;
			}
			if (!(ret_flags & BLOCK_CHANGED))
				continue;

#if 0

			/* FIXME: Check for overflow! */
			ex->ee_block = extent.e_pblk;
			ex->ee_start = extent.ee_start & 0xFFFFFFFF;
			ex->ee_start_hi = extent.ee_start >> 32;
			ex->ee_len = extent.e_len;

			if (ex_prev &&
			    (extent.e_pblk ==
			     extent_prev.e_pblk + extent_prev.e_len) &&
			    (extent.e_lblk + j ==
			     extent_prev.e_lblk + extent_prev.e_len) &&
			    (extent.e_len + extent_prev.e_len < 65535)) {
				extent_prev.e_len += extent.e_len;
				ex_prev->ee_len += extent.e_len;
				ex->ee_len = 0;
				extent.e_len = 0;
			}
			if (ex->ee_len == 0) {
				extent_remove(eh, ex);
				i--; ex--;
			} else {
				ex_prev = ex;
				extent_prev = extent;
			}
			ret |= BLOCK_CHANGED;
#endif
		}
	} else {
		char *block_buf;
		struct ext3_extent_idx *ix;
		struct ext3_extent_header *nh;

		ret = ext2fs_get_mem(fs->blocksize, &block_buf);
		if (ret)
			return ret;

		ext_show_header(eh);
		ix = EXT_FIRST_INDEX(eh);
		for (i = 0; i < eh->eh_entries; i++, ix++) {
			ext_show_index(ix);
			if (!(flags & BLOCK_FLAG_DEPTH_TRAVERSE) &&
			    !(flags & BLOCK_FLAG_DATA_ONLY)) {
				ret |= (*meta_func)(fs, ix->ei_leaf,
						    -1, 0, priv_data);
				if (ret & BLOCK_ABORT)
					goto free_buf;
			}

			retval = ext2fs_read_ext_block(fs, ix->ei_leaf,
						       block_buf);
			if (retval) {
				ret |= BLOCK_ERROR;
				goto free_buf;
			}
			nh = (struct ext3_extent_header *)block_buf;
			if (ext2fs_extent_header_verify(nh, fs->blocksize)){
				ret |= BLOCK_ERROR;
				goto free_buf;
			}

			ret_flags = block_iterate_extents(fs, nh,
							  ix->ei_leaf,
							  flags,
							  func, meta_func,
							  priv_data);
			if (ret_flags & BLOCK_CHANGED)
				retval =
					ext2fs_write_ext_block(fs,
							       ix->ei_leaf,
							       block_buf);
			if (ret_flags & BLOCK_ABORT) {
				ret |= BLOCK_ABORT;
				goto free_buf;
			}
#if 0
			if ((flags & BLOCK_FLAG_DEPTH_TRAVERSE) &&
			    !(flags & BLOCK_FLAG_DATA_ONLY))
				ret |= (*func)(fs, &ix->ei_leaf,
					       BLOCK_COUNT_IND, ref_block,
					       i, ctx->priv_data);
#endif
		}

	free_buf:
		ext2fs_free_mem(&block_buf);
	}
	return ret;
}


errcode_t ext2fs_extent_iterate(ext2_filsys fs,
				ext2_ino_t ino,
				int	flags,
				int (*func)(ext2_filsys fs,
					    struct ext2fs_extent *extent,
					    void	*priv_data),
				int (*meta_func)(ext2_filsys fs,
						 blk64_t blk,
						 int blk_type,
						 char *buf,
						 void *priv_data),
				void *priv_data)
{
	struct ext2_inode inode;
	errcode_t errcode;
	struct ext3_extent_header *eh;
	struct ext3_extent_idx *ix;
	int		i, ret;

	EXT2_CHECK_MAGIC(fs, EXT2_ET_MAGIC_EXT2FS_FILSYS);

	errcode = ext2fs_read_inode(fs, ino, &inode);
	if (errcode)
		return errcode;

	if (!(inode.i_flags & EXT4_EXTENTS_FL)) {
		/*
		 * XXX not yet....
		 */
		return EOPNOTSUPP;
	}

	eh = (struct ext3_extent_header *) inode.i_block;

	printf("I'm in ext2fs_extent_iterate!\n");
	ext_show_header(eh);

	ix = EXT_FIRST_INDEX(eh);
	for (i = 0; i < eh->eh_entries; i++, ix++) {
		ext_show_index(ix);
	}

	ret = block_iterate_extents(fs, eh, -1, flags, func, meta_func,
				    priv_data);

	return 0;
}


#ifdef DEBUG

int test_iter_func(ext2_filsys fs, struct ext2fs_extent *extent,
		   void *priv_data)
{
	printf("extent_iter: l_blk %lld, len %d, p_blk %lld\n",
	       extent->e_lblk, extent->e_len, extent->e_pblk);
}

int test_iter_meta_func(ext2_filsys fs, blk64_t blk, int blk_type,
		    char *buf, void *priv_data)
{
	printf("extent_meta_iter: blk %lld, type %d\n", blk, blk_type);
	return (0);
}

main(int argc, char **argv)
{
	ext2_filsys	fs;
	errcode_t	retval;
	ext2_ino_t ino = 17;
	const char *	fn = "extents_image";

	if (argc >= 2)
		ino = atoi(argv[1]);

	if (argc >= 3)
		fn = argv[2];

	initialize_ext2_error_table();
	retval = ext2fs_open ("extents_image", EXT2_FLAG_SOFTSUPP_FEATURES,
			      0, 0, unix_io_manager, &fs);
	if (retval) {
		com_err("tst_extent", retval,
			 "while trying to open %s", fn);
		exit(1);
	}

	retval = ext2fs_extent_iterate(fs, ino, 0, test_iter_func,
				       test_iter_meta_func, 0);
	if (retval) {
		com_err("tst_extent", retval,
			 "while trying to interate over extents");
		exit(1);
	}


	exit(0);
}
#endif
